import logging
from typing import Union, Optional, Callable

from . import _client

logger = logging.getLogger(__name__)


def initialize(
    *,
    host: str,
    port: Optional[Union[str, int, None]] = None,
    username: str,
    password: str,
    validate_cert: Optional[Union[bool, None]] = True,
    use_ssl: Optional[Union[bool, None]] = True,
) -> Callable:
    """ Initialize a WinRM session and get a connector context manager.

    :param host: The hostname or IP address of the Windows Server to connect to.
    :param port: The listening WinRM port. Defaults to 5986 if `use_ssl=True` and 5985 if `use_ssl=False`.
    :param username: The username for authentication.
    :param password: The password for authentication.
    :param use_ssl: Denotes whether to connect over HTTP or HTTPS. Defaults to `True` (HTTPS).
    :param validate_cert: Denotes whether to attempt server certificate validation. Defaults to `True`.

    :raises exceptions.CertificateError: if certificate verification failed.
    :raises exceptions.NetworkError: if the network connection failed.

    :returns: A context manager that yields an open WinRM connection.

    The returned context manager should only be called using a `with statement
    <https://docs.python.org/3.8/reference/compound_stmts.html#with>`_.

    All Windows Server operations must happen inside the `with` block. You do not need to know how to
    open or close a WinRM - the context manager handles all of this for you.

    You can re-use an initialized session as often as necessary. This makes it easier to separate
    offline programming logic from Windows Server module operations as much as possible.

    See the example below, or the more complete :ref:`index:Example Script` to view it in action.

    .. code-block:: python

        # intialize a session
        session = client.initialize(
            host='host',
            username='username',
            password='password',
            use_ssl=True
        )

        with session() as connection:
            # run some Powershell
            client.run_powershell(connection, script="echo $env:computername")

    .. seealso::

        `Powershell Remoting
        <https://docs.microsoft.com/en-us/powershell/scripting/learn/remoting/winrmsecurity?view=powershell-6\
#powershell-remoting-default-settings>`_

    """

    if not port:
        port = 5986 if use_ssl else 5985
    return _client._get_session(
        host=host, port=port, username=username, password=password, validate_cert=validate_cert, use_ssl=use_ssl
    )


def run_dos(connection, *, command: str, environment: Optional[dict] = None) -> str:
    """ Run an arbitrary DOS command on a Windows Server.

    :param connection: An open WinRM connection.
    :param command: The DOS command to run.
    :param environment: An optional dictionary containing environment keys and values to set on the executing script.

    :raises exceptions.DOSCommandError: if the command errored on the server.

    :returns: A string containing the command output.

    Example:

    .. code-block:: python

        # intialize a session
        session = client.initialize(
            host='host',
            username='username',
            password='password',
            use_ssl=True
        )

        with session() as connection:
            # get the server versionm
            client.run_dos(connection, command="ver")

    """
    return _client._run_dos(connection, command=command, environment=environment)


def run_powershell(
    connection, *, script: str, configuration_name: str = "Microsoft.PowerShell", environment: Optional[dict] = None
) -> str:
    """ Run an arbitrary Powershell script on a Windows Server.

    :param connection: An open WinRM connection.
    :param script: The Powershell script. Can be multiline.
    :param configuration_name: An optional PowerShell configuration endpoint to use when executing the script.
    :param environment: An optional dictionary containing environment keys and values to set on the executing script.

    :raises exceptions.PowershellError: if the script errored on the server.

    :returns: A string containing the command output.

    Example:

    .. code-block:: python

        # intialize a session
        session = client.initialize(
            host='host',
            username='username',
            password='password',
            use_ssl=True
        )

        with session() as connection:
            # get the computername
            client.run_powershell(connection, script="echo $env:computername")

    """
    return _client._run_powershell(
        connection, script=script, configuration_name=configuration_name, environment=environment
    )


def upload_file(connection, *, source: str, destination: str, configuration_name: str = "Microsoft.PowerShell") -> str:
    """ Copy a file from the PPA task container to a Windows Server. This function
    will overwrite any existing file matching the destination path.

    :param connection: An open WinRM connection.
    :param source: The local path of the file to copy.
    :param destination: The destination path on the Windows Server.
    :param configuration_name: An optional PowerShell configuration endpoint to use when uploading the file.

    :raises exceptions.PathError: if the source or destination path is invalid.
    :raises exceptions.CopyFileError: if the file could not be copied to the Windows Server.

    :returns: A string containing the absolute path of the remote copied file.

    Example:

    .. code-block:: python

        # intialize a session
        session = client.initialize(
            host='host',
            username='username',
            password='password',
            use_ssl=True
        )

        with session() as connection:
            # copy a local log file to the server and store the remote path
            remote_path = client.upload_file(
                connection,
                source="./log.txt",
                destination="C:\\Users\\Administrator\\log.txt"
            )

    """
    return _client._upload_file(
        connection, source=source, destination=destination, configuration_name=configuration_name
    )


def download_file(connection, *, source: str, destination: str, configuration_name: str = "Microsoft.PowerShell"):
    """ Download a file from a Windows Server and saves it in the PPA task
    container. This function will overwrite any existing file matching the
    destination path.

    :param connection: An open WinRM connection.
    :param source: The path of the file to download from the Windows Server.
    :param destination: The destination path in the PPA task container.
    :param configuration_name: An optional PowerShell configuration endpoint to use when downloading the file.

    :raises exceptions.PathError: if the source or destination path is invalid.
    :raises exceptions.CopyFileError: if the file could not be copied to the Windows Server.

    Example:

    .. code-block:: python

        # intialize a session
        session = client.initialize(
            host='host',
            username='username',
            password='password',
            use_ssl=True
        )

        with session() as connection:
            # download a local log file from the server
            client.download_file(
                connection,
                source="C:\\Users\\Administrator\\log.txt",
                destination="./log.txt",
            )

    """
    return _client._download_file(
        connection, source=source, destination=destination, configuration_name=configuration_name
    )
