import logging
import functools
import json
from collections import namedtuple
from typing import List, NamedTuple

from . import client

logger = logging.getLogger(__name__)

_header = ["Name", "Display Name", "State", "Status", "Account", "Description"]
_fields = ["name", "displayname", "state", "status", "startname", "description"]

# The field names returned by Powershell - these are only
# used when running the powershell command, and loading
# the records into namedtuples.
_powershell_return_fields = ["Name", "DisplayName", "State", "Status", "StartName", "Description"]
_powershell_field_template = ", ".join(_powershell_return_fields)

_windows_service = namedtuple("windows_service", _fields)  # type: ignore


def _as_objects(func):
    @functools.wraps(func)
    def _fn(connection, *args, **kwargs) -> List[NamedTuple]:
        output = func(connection, *args, **kwargs)
        if not output:
            return []
        json_data = json.loads(output)
        if not isinstance(json_data, list):
            json_data = [json_data]
        return [
            _windows_service(*[service.get(field, "") for field in _powershell_return_fields])  # type: ignore
            for service in json_data
        ]

    return _fn


@_as_objects
def get_all(connection) -> List[NamedTuple]:
    """ Get all Windows Services.

    :returns: A list of :ref:`services:Service Records`.

    Example:

    .. code-block:: python

        with session() as connection:
            all_services = services.get_all(connection)

    """
    return client.run_powershell(  # type: ignore
        connection, script=f"gwmi win32_service | select {_powershell_field_template} | ConvertTo-Json"
    )


@_as_objects
def get_by_name(connection, *, name: str) -> List[NamedTuple]:
    """ Get a Windows Service using a name.

    :param name: The name of the service (must be an exact match).

    :returns: A list of :ref:`services:Service Records`.

    Example:

    .. code-block:: python

        with session() as connection:
            spooler_service = services.get_by_name(connection, name="spooler")

    """
    return client.run_powershell(  # type: ignore
        connection,
        script=f"gwmi win32_service -Filter \"name='{name}'\" | "
        f"select {_powershell_field_template} | ConvertTo-Json",
    )


@_as_objects
def search_by_name(connection, *, name: str) -> List[NamedTuple]:
    """ Get one or more Windows Services by searching with a name.

    :param name: The name of the service. Can be a substring. Leading and trailing
        wildcards are added to the name before the command is run.

    :returns: A list of :ref:`services:Service Records`.

    Example:

    .. code-block:: python

        with session() as connection:
            search_results = services.search_by_name(connection, name="pool")

    """
    return client.run_powershell(  # type: ignore
        connection,
        script=f"gwmi win32_service -Filter \"name like '%{name}%'\" | "
        f"select {_powershell_field_template} | ConvertTo-Json",
    )


def restart(connection, *, name: str) -> None:
    """ Restart a Windows Service.

    :param name: The name of the service to restart. Must be an exact match.

    Example:

    .. code-block:: python

        with session() as connection:
            services.restart(connection, name="spooler")

    """
    client.run_powershell(connection, script=f'Restart-Service -name "{name}"')
